import sys
import ast
from gurobipy import *
import time

debug_output = False

class Node:
    requests = []
    r_requests = []

    def __init__(self, w_requests, length=0, r_requests=None):
        self.requests = w_requests
        self.cs_len = length
        self.r_requests = r_requests

    def nodePrint(self):
        print("Print node with " + str(len(self.requests)) + " requests...")
        for r in self.requests:
            print(r + ", ")
        print(".... and duration " + str(self.cs_duration))

class MinColorInstance:
    nodes = []

    def __init__(self, nodes=[], cslenInst=False, rwInst=False, filename="temp.txt"):
        self.nodes = nodes
        self.filename = filename
        self.fileidx = 0;
        self.n_nodes = len(nodes)
        self.cslen_instance = cslenInst
        self.rw_instance = rwInst

    def read(self):
        # print "Filename is: " + self.filename
        f = open(self.filename, "r")
        str_line = []
        self.n_nodes = 0
        for line in f:
            str_line.append(line)
            self.n_nodes += 1
            reqs_cslen = ast.literal_eval(line)
            if isinstance(reqs_cslen[0], list):
                if isinstance(reqs_cslen[0][0], list):
                    self.rw_instance = True
                    temp_node = Node(reqs_cslen[0][1], reqs_cslen[1], reqs_cslen[0][0])
                    self.nodes.append(temp_node)
                else:
                    self.cslen_instance = True
                    temp_node = Node(reqs_cslen[0], reqs_cslen[1])
                    self.nodes.append(temp_node)
            else:
                temp_node = Node(reqs_cslen, -1)
                self.nodes.append(temp_node)

    def setFileIdx(self):
        print("Setting fileidx... NOT IMPLEMENTED")

    def setFileName(self):
        self.filename = "./input/"
        if self.fileidx < 10:
            self.filename += "00"
        elif self.fileidx < 100:
            self.filename += "0"
        self.filename += str(self.fileidx) + ".txt"

    def instancePrint(nodes):
        print("N nodes: " + str(len(nodes)))
        for n in nodes:
            print(n)

    def solve_greedy(self):

        start_time = time.time()

        colors = []
        for idx in range(self.n_nodes):
            colors.append(-1)
        if self.rw_instance is False:
            for idx in range(self.n_nodes):
                assigned_colors = []
                req1 = self.nodes[idx].requests
                for idy in range(self.n_nodes):
                    req2 = self.nodes[idy].requests
                    v_intersection = [value for value in req1 if value in req2]
                    if v_intersection != [] and colors[idy] != -1:
                        if colors[idy] not in assigned_colors:
                            assigned_colors.append(colors[idy])
                assigned_colors.sort()

                next_color = 1
                for idy in range(len(assigned_colors)):
                    if (idy+1) != assigned_colors[idy]:
                        break
                    next_color = next_color + 1

                colors[idx] = next_color
        else:
            for idx in range(self.n_nodes):
                assigned_colors = []
                w_req1 = self.nodes[idx].requests
                r_req1 = self.nodes[idx].r_requests
                for idy in range(self.n_nodes):
                    w_req2 = self.nodes[idy].requests
                    r_req2 = self.nodes[idy].r_requests
                    v_intersection_ww = [value for value in w_req1 if value in w_req2]
                    v_intersection_rw = [value for value in r_req1 if value in w_req2]
                    v_intersection_wr = [value for value in w_req1 if value in r_req2]
                    if (v_intersection_ww != [] or v_intersection_rw != [] or v_intersection_wr != []) and colors[idy] != -1:
                        if colors[idy] not in assigned_colors:
                            assigned_colors.append(colors[idy])
                assigned_colors.sort()

                next_color = 1
                for idy in range(len(assigned_colors)):
                    if (idy+1) != assigned_colors[idy]:
                        break
                    next_color = next_color + 1

                colors[idx] = next_color

        greedy_time = time.time() - start_time
        return [greedy_time, colors]

def solve(inst, n_colors, timeout = 30.0):

    start_setup = time.time()

    model = Model("ColorFinder")
    model.setParam(GRB.Param.OutputFlag, 0)

    n_var = 0
    n_con = 0

    color = []
    duration = []
    if inst.cslen_instance is False and inst.rw_instance is False:
        for i in range(n_colors):
            temp_var = model.addVar(lb=0.0,ub=1.0,obj=0.0,vtype=GRB.BINARY,name="w["+str(i)+"]")
            color.append(temp_var)
        model.update()
    else:
        for i in range(n_colors):
            temp_var = model.addVar(lb=0.0,ub=1000.0,obj=0.0,vtype=GRB.INTEGER,name="dur["+str(i)+"]")
            duration.append(temp_var)
        model.update()

    x = []
    for idx in range(inst.n_nodes):
        x_t = []
        for idz in range(n_colors):
            temp_var = model.addVar(lb=0.0,ub=1.0,obj=0.0,vtype=GRB.BINARY,name="x["+str(idx)+"]["+str(idz)+"]")
            x_t.append(temp_var)
        x.append(x_t)
#    print x

    if inst.rw_instance is False:
        for idx in range(inst.n_nodes):
            req1 = inst.nodes[idx].requests
            for idy in range(idx+1, inst.n_nodes):
                req2 = inst.nodes[idy].requests
                v_intersection = [value for value in req1 if value in req2]
                if v_intersection != []:
                    for idz in range(n_colors):
                        model.addConstr(x[idx][idz] + x[idy][idz] <= 1)
    else:
        for idx in range(inst.n_nodes):
            w_req1 = inst.nodes[idx].requests
            r_req1 = inst.nodes[idx].r_requests
            for idy in range(idx+1, inst.n_nodes):
                w_req2 = inst.nodes[idy].requests
                r_req2 = inst.nodes[idy].r_requests
                v_intersection_ww = [value for value in w_req1 if value in w_req2]
                v_intersection_rw = [value for value in r_req1 if value in w_req2]
                v_intersection_wr = [value for value in w_req1 if value in r_req2]
                if v_intersection_ww != [] or v_intersection_rw != [] or v_intersection_wr != []:
                    for idz in range(n_colors):
                        model.addConstr(x[idx][idz] + x[idy][idz] <= 1)


    if inst.cslen_instance is False and inst.rw_instance is False:
        for idx in range(inst.n_nodes):
            xexpr = LinExpr(0.0)
            for idz in range(n_colors):
                xexpr += x[idx][idz]
                model.addConstr(x[idx][idz] <= color[idz])
            model.addConstr(xexpr <= 1)
            model.addConstr(xexpr >= 1)

        objective = LinExpr(0.0)
        for idz in range(n_colors):
            objective += color[idz]

    else:
        for idx in range(inst.n_nodes):
            cs_len = inst.nodes[idx].cs_len
            xexpr = LinExpr(0.0)
            for idz in range(n_colors):
                xexpr += x[idx][idz]
                model.addConstr(duration[idz] >= cs_len*x[idx][idz])
            model.addConstr(xexpr <= 1)
            model.addConstr(xexpr >= 1)
        objective = LinExpr(0.0)

        for idz in range(n_colors):
            objective += duration[idz]

    model.setObjective(objective, GRB.MINIMIZE)

    model.setParam(GRB.Param.Threads, 1)
    model.setParam(GRB.Param.TimeLimit, timeout)

    end_setup = time.time()
    setup_time = end_setup - start_setup

    model.optimize()

    solve_time = time.time() - end_setup

    status = model.Status
    if (status == GRB.TIME_LIMIT):
        print("Optimization stopped due to time limit")
    if (status != GRB.OPTIMAL) or (status == GRB.INF_OR_UNBD) or (status == GRB.INFEASIBLE) or (status == GRB.UNBOUNDED):
        print("Optimization stopped with status " + str(status))

    #else:
    #    print("Bounds:\t" + str(int(model.objVal)) + "\t" + str(int(model.objBound)))

    if debug_output is True:
        outfile = "./debug_output.txt"
        of = open(outfile, "w")
        of.write("\t" + str(int(model.objVal)) + "\n")
        of.write("Colors\n")
        for idx in range(inst.n_nodes):
            for idz in range(n_colors):
                c_val = round(x[idx][idz].getAttr(GRB.Attr.X))
                if c_val == 1:
                    of.write(str(idz+1) + "\n")
                    break
       # if inst.cslen_instance is False and inst.rw_instance is False:
       #     for idz in range(n_colors):
       #         of.write("\t" + str(int(0+color[idz].x)) + "\n")
       # else:
       #     for idz in range(n_colors):
       #         of.write("\t" + str(int(0+duration[idz].x)) + "\n")
        of.write("\nNodes\n")
        for idx in range(inst.n_nodes):
            for idz in range(n_colors):
                of.write("\t" + str(round(x[idx][idz].getAttr(GRB.Attr.X))) + "\n")
            of.write("\n")
        of.close()
        return [-1,-1,-1]
    #elif inst.cslen_instance is False:
    else:
        if (status == GRB.TIME_LIMIT):
            return [setup_time,-1,[]]
        colors = []
        for idx in range(inst.n_nodes):
            color_found = False
            for idz in range(n_colors):
                c_val = round(x[idx][idz].getAttr(GRB.Attr.X))
                if c_val == 1:
                    color_found = True
                    colors.append(idz+1)
                    break
            if color_found != True:
                print("Error: no color was assigned for idx: " + str(idx))
        return [setup_time, solve_time, colors]
   # else:
   #     outfile = "./duration_output.txt"
   #     of = open(outfile, "w")
   #     for idx in range(n_colors):
   #         of.write(str(duration[idx].getAttr(GRB.Attr.X)) + "\n")
   #     of.close()


if __name__ == "__main__":
    if len(sys.argv) - 1 < 2:
        print("Requires two arguments: filename num_colors")
    filename = sys.argv[1]
    n_colors = int(sys.argv[2])

    inst = MinColorInstance(filename=filename)
    inst.read()
    [setup_time, solve_time, colors] = inst.solve(n_colors)
    [exec_time, colors] = inst.solve_greedy()
    print(colors)
