#include "sharedres.h"
#include "blocking.h"
#include "rw-blocking.h"
#include <iostream>

#include "stl-helper.h"

BlockingBounds* group_mcs_bounds(const ResourceSharingInfo& info,
				 unsigned int procs_per_cluster,
				 int dedicated_irq)
{
	// split everything by partition
	Clusters clusters;

	split_by_cluster(info, clusters);

	// split each partition by resource
	ClusterResources resources;

	split_by_resource(clusters, resources);

	// sort each contention set by request length
	sort_by_request_length(resources);

        std::vector<const RequestBound*> write_reqs;
        unsigned int k = 0;
        for(k=0; k<resources[0].size(); k++)
        {
            foreach(resources[0][k], it){
                const RequestBound* req = *it;
                unsigned int m = 0;
                bool req_already_recorded = false;
                for(m=0; m<write_reqs.size(); m++){
                    if(write_reqs[m]->get_task() == req->get_task()){
                        req_already_recorded = true;
                        break;
                    }
                }
                if (req_already_recorded == false){
                    write_reqs.push_back(req);
                }
            }
        }
        sort_by_request_length(write_reqs);

//std::cout << "-Writes-\n";
//foreach(write_reqs, it)
//{
//    const RequestBound* req = *it;
//    std::cout <<req->get_request_length() << "\n";
//}
//std::cout << "---\n";

	// We need for each task the maximum request span.  We also need the
	// maximum direct blocking from remote partitions for each request. We
	// can determine both in one pass.

	unsigned int i;

	// direct blocking results
	BlockingBounds* _results = new BlockingBounds(info);
	BlockingBounds& results = *_results;

	for (i = 0; i < info.get_tasks().size(); i++)
	{
		const TaskInfo& tsk  = info.get_tasks()[i];
		Interference bterm;

                if (tsk.get_requests().size() > 0)
		{
                        const RequestBound* this_req = &tsk.get_requests()[0];

			Interference wblocking, wblocking_w1;

                        int idx;
                        unsigned int parallelism = procs_per_cluster;
	                enumerate(resources, ct, idx)
		        {
                            parallelism = procs_per_cluster;

		            if (idx == dedicated_irq)
			        parallelism--;

		            if (parallelism && (int) tsk.get_cluster() == idx)
			        parallelism--;
		        }
			// Calculate wblocking
                        if (parallelism == 0)
                        {
                            wblocking = 0;
                        }
                        else
                        {
                                unsigned int max_requests_per_source = 1;
                                unsigned int count = 0;
                                unsigned int total_length = 0;
                                    unsigned int remaining = parallelism;
                                    foreach(write_reqs, it)
                                    {
                                       const RequestBound* req = *it;
                                       if (!remaining)
                                           break;
                                       if (req->get_task() != &tsk)
                                       {
                                           unsigned int num;
                                           num = std::min(req->get_max_num_requests(tsk.get_response()),
                                                  max_requests_per_source);
                                           num = std::min(num, remaining);
                                           total_length += num*req->get_request_length();
                                           count += num;
                                           remaining -= num;
                                       }
                                    }
                                    wblocking.total_length = total_length;
                                    wblocking.count = count;

                        }

				wblocking_w1 = wblocking;
				wblocking_w1.total_length += this_req->get_request_length();
				wblocking_w1.count    += 1;

			    // combine
			    results.raise_request_span(i, wblocking_w1);
			    bterm += wblocking;
		}
		results[i] = bterm;
	} // foreach

	// This is the initial delay due to priority donation.
	charge_arrival_blocking(info, results);

	return _results;
}

